import { h } from "@stencil/core";
import { animate } from "animejs";
import { funnelDefs, colorFor } from "../../utils";
let barHeight = [];
let barRefs = undefined;
const bgdFill = colorFor('white');
const xFill = colorFor('bada55');
let tween = undefined;
let tween2 = undefined;
const processColor = (next) => {
    if (!barRefs)
        return;
    // IMPORTANT:  we are tweening between the hex values, not the names...
    if (tween) {
        tween.pause();
    }
    tween = animate(barRefs, {
        fill: colorFor(next),
        ease: 'inOutCubic',
    });
};
const processValues = (next) => {
    if (!barRefs)
        return;
    if (tween2) {
        tween2.pause();
    }
    // NOTE:  next values between 0-100 (aka. percentage)
    tween2 = animate(barRefs, {
        y: (_bar, indx, _length) => barHeight[indx] * ((100 - next[indx]) / 100),
        ease: 'inOutCubic',
    });
};
const createFunnel = ({ id, height, width, x, y, drop }) => {
    const obj = {
        id: id,
        height,
        width,
        x,
        y,
        maskId: `mask-${id}`,
        maskUri: `url('#mask-${id}')`,
        mask: `0.5,0.5,0.5,${height - 0.5},${width - 0.5},${height - drop - 0.5},${width - 0.5},${drop + 0.5}`,
        matrix: `matrix(1,0,0,1,${x},${y})`,
        points: `0,0,0,${height},${width},${height - drop},${width},${drop}`,
    };
    return obj;
};
const generateBars = () => {
    return funnelDefs.map(item => {
        return createFunnel(item);
    });
};
const FunnelGfx = _props => {
    const bars = generateBars();
    const stroke = 'currentColor';
    // IMPORTANT:  initialize these arrays when you render...
    barHeight = Array(bars.length);
    barRefs = Array(bars.length);
    return (h("div", { class: "funnel-gfx" }, h("svg", { viewBox: "0 0 600 420" }, h("desc", null, "Created with Snap"), h("defs", null, bars.map(item => {
        return (h("mask", { id: item.maskId }, h("polygon", { points: item.mask, fill: bgdFill })));
    })), bars.map((item, index) => {
        const { height, width, maskUri, matrix, points } = item;
        barHeight[index] = height;
        return (h("g", { transform: matrix }, h("polygon", { points: points, fill: bgdFill, stroke: stroke }), h("rect", { ref: el => (barRefs[index] = el), x: "0", y: height, width: width, height: height, mask: maskUri, fill: xFill, stroke: stroke })));
    }))));
};
export { FunnelGfx, processColor, processValues };
export default FunnelGfx;
